import logging
from itertools import count
from typing import List
from typing import Optional

import requests

from socorepo.consts import CLF_SOURCES
from socorepo.locators.github_tags import GitHubTags, GitHubBasicAuth, GitHubOAuthToken, GitHubOAuthApp, \
    GitHubTagsComponentData
from socorepo.locators.helpers import fetch_json
from socorepo.structs import ComponentPrototype, AssetPrototype

log = logging.getLogger("socorepo")


def fetch_component_prototypes(locator: GitHubTags) -> List[ComponentPrototype]:
    json_tag_listing = fetch_json_from_endpoint(locator, endpoint="tags")
    json_release_listing = fetch_json_from_endpoint(locator, endpoint="releases")

    components = []

    for json_tag in json_tag_listing:
        json_release = next((jr for jr in json_release_listing if jr["tag_name"] == json_tag["name"]), None)
        components.append(_parse_component(locator, json_tag, json_release))

    return components


def fetch_json_from_endpoint(locator: GitHubTags, endpoint: str) -> list:
    api_url = f"https://api.github.com/repos/{locator.owner}/{locator.repository}/{endpoint}?"
    if isinstance(locator.auth, GitHubOAuthApp):
        api_url += f"client_id={locator.auth.client_id}&client_secret={locator.auth.client_secret}"
    api_url += "per_page=100&page="

    session = requests.Session()
    session.headers.update({"Accept": "application/vnd.github.v3+json"})
    if isinstance(locator.auth, GitHubBasicAuth):
        session.auth = (locator.auth.username, locator.auth.password)
    elif isinstance(locator.auth, GitHubOAuthToken):
        session.headers.update({"Authorization": f"token {session.auth.oauth_token}"})

    json_listing = []

    for page in count(1):
        try:
            json_resp = fetch_json(url=(api_url + str(page)), session=session)
        except IOError as e:
            log.error("Cannot fetch listing of %s: %s", endpoint, e)
            return []

        if not isinstance(json_resp, list):
            log.error("GitHub returned an invalid JSON response that is not a listing of %s. "
                      "Are you sure that the repository owner '%s' and repository name '%s' actually exist and "
                      "the (optional) authentication information is correct? "
                      "First 100 chars of server's response are: %.100s",
                      endpoint, locator.owner, locator.repository, json_resp)
            return []

        # Stop when we're past the last page.
        if len(json_resp) == 0:
            break

        json_listing += json_resp

    return json_listing


def _parse_component(locator: GitHubTags, json_tag: dict, json_release: Optional[dict]) -> ComponentPrototype:
    version = json_tag["name"]
    commit = json_tag["commit"]["sha"]

    assets = [
        AssetPrototype(filename=f"{locator.owner}-{locator.repository}-{version}.zip",
                       file_size=None,
                       url=json_tag["zipball_url"],
                       checksums={},
                       forced_clfs=[CLF_SOURCES]),
        AssetPrototype(filename=f"{locator.owner}-{locator.repository}-{version}.tar.gz",
                       file_size=None,
                       url=json_tag["tarball_url"],
                       checksums={},
                       forced_clfs=[CLF_SOURCES])
    ]

    if json_release:
        for json_asset in json_release["assets"]:
            assets.append(AssetPrototype(filename=json_asset["name"],
                                         file_size=json_asset["size"],
                                         url=json_asset["browser_download_url"],
                                         checksums={}))

    return ComponentPrototype(version=version, assets=assets,
                              extra_data=GitHubTagsComponentData(commit=commit))



from dataclasses import dataclass
from typing import List, Dict
from urllib.parse import urljoin

from markupsafe import Markup

from socorepo.config import TomlDict
from socorepo.locators.helpers import ensure_trailing_slash
from socorepo.structs import Locator, ComponentPrototype, Component


@dataclass(frozen=True)
class PyPI(Locator):
    server: str
    verify_tls_certificate: bool
    project: str

    def fetch_component_prototypes(self) -> List[ComponentPrototype]:
        from . import fetcher  # avoid cyclic dependencies
        return fetcher.fetch_component_prototypes(self)

    def component_info_table(self, component: Component) -> Dict[str, Markup]:
        project_url = urljoin(self.server, self.project)

        return {
            "Component host": Markup("PyPI package index"),
            "PyPI server": Markup(f'<a href="{self.server}" target="_blank">{self.server}</a>'),
            "PyPI project": Markup(f'<a href="{project_url}" target="_blank">{self.project}</a>')
        }


def parse_locator(toml_locator: TomlDict):
    return PyPI(server=ensure_trailing_slash(toml_locator.req("server", str)),
                verify_tls_certificate=toml_locator.opt("verify_tls_certificate", bool, fallback=True),
                project=toml_locator.req("project", str))



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Optional, List, Dict, Pattern

from markupsafe import Markup


# ======================
# === Config objects ===
# ======================

@dataclass(frozen=True)
class Project:
    id: str
    label: str
    description: Optional[Markup]
    excluded_asset_clfs: List[str]
    featured_asset_type_matchers: List[AssetTypeMatcher]
    locator: Locator


class Locator(ABC):
    @abstractmethod
    def fetch_component_prototypes(self) -> List[ComponentPrototype]:
        raise NotImplementedError

    def component_info_table(self, component: Component) -> Dict[str, Markup]:
        raise NotImplementedError


@dataclass(frozen=True, order=True)
class VersionQualifier:
    ordinal: int  # has to be the first attribute so it influences sorting
    name: str
    color: str
    version_section_regex: Optional[Pattern]  # present if and only if default = False
    default: bool
    stable: bool

    def matches(self, version_section: str) -> bool:
        return self.version_section_regex and self.version_section_regex.search(version_section) is not None


@dataclass(frozen=True)
class AssetClfMatcher:
    clf: str
    filename_regex: Pattern

    def matches(self, version: str) -> bool:
        return self.filename_regex.search(version) is not None


@dataclass(frozen=True)
class AssetTypeMatcher:
    pattern: str
    _req_clf_regexes: List[Pattern] = field(default_factory=list, init=False, repr=False, compare=False)
    _opt_clf_regexes: List[Pattern] = field(default_factory=list, init=False, repr=False, compare=False)

    def __post_init__(self):
        for clf_pattern in self.pattern.split(" "):
            req = True
            if clf_pattern.startswith("?"):
                req = False
                clf_pattern = clf_pattern[1:]

            if "?" in clf_pattern:
                raise ValueError(f"While parsing pattern {self.pattern}: The asset clf {clf_pattern} contains '?' "
                                 f"at a position that is not the beginning of the string. This is disallowed.")

            regex = re.compile(re.escape(clf_pattern).replace(r"\*", ".*").replace(r"\|", "|"), re.IGNORECASE)
            (self._req_clf_regexes if req else self._opt_clf_regexes).append(regex)

    def matches(self, type_: AssetType):
        matched_regexes = set()
        matched_clfs = set()

        for regex in self._req_clf_regexes + self._opt_clf_regexes:
            for clf in type_.clfs:
                if regex.match(clf):
                    matched_regexes.add(regex)
                    matched_clfs.add(clf)

        # Returns true iff every required patterns has matched with a clf and every clf has matched with any pattern.
        return len(set(self._req_clf_regexes).difference(matched_regexes)) == 0 and matched_clfs == set(type_.clfs)


# =====================
# === Fetched stuff ===
# =====================

@dataclass(frozen=True)
class ComponentPrototype:
    version: str
    assets: List[AssetPrototype]
    extra_data: Any = None  # optional


@dataclass(frozen=True)
class AssetPrototype:
    filename: str
    file_size: Optional[int]  # in bytes
    url: str
    # compare=False implies hash=False and that's necessary as lists and dicts aren't easily hashable.
    # Instances of this class need to be hashable because the fetcher procedure puts them in a dict.
    checksums: Dict[str, str] = field(compare=False)
    forced_clfs: List[str] = field(default_factory=list, compare=False)


@dataclass(frozen=True)
class Component:
    version: str
    qualifier: VersionQualifier
    assets: List[Asset]
    extra_data: Any = None  # optional


@dataclass(frozen=True)
class Asset:
    filename: str
    file_size: Optional[int]  # in bytes
    url: str
    checksums: Dict[str, str]
    type: AssetType
    featured: bool
    matcher_causing_featuring: Optional[AssetTypeMatcher]


@dataclass(frozen=True)
class AssetType:
    clfs: List[str]

    def __str__(self):
        return " ".join(self.clfs)



from json import JSONDecodeError
from urllib.parse import urljoin

import requests


def ensure_trailing_slash(url: str):
    return urljoin(url + "/", ".")


def fetch_json(*, url: str, session: requests.Session = None, verify: bool = True):
    # Any IOError raised here will just be raised to the calling function, which is desired behavior.
    if session:
        resp = session.get(url, verify=verify)
    else:
        resp = requests.get(url, verify=verify)

    try:
        return resp.json()
    except JSONDecodeError as e:
        raise IOError("Cannot decode JSON response. "
                      f"First 100 chars of server's response are: {resp.text:.100}\n"
                      f"Exception is: {e}")



import logging
import os
from typing import List
from urllib.parse import urljoin

from socorepo.locators.helpers import fetch_json
from socorepo.locators.nexus3 import Nexus3, Nexus3ComponentData
from socorepo.structs import ComponentPrototype, AssetPrototype

log = logging.getLogger("socorepo")


def fetch_component_prototypes(locator: Nexus3) -> List[ComponentPrototype]:
    if not locator.verify_tls_certificate:
        log.warning("As specified in the repository settings, will now connect to the HTTPS server '%s' without"
                    " verifying its TLS certificate.", locator.server)

    api_url = urljoin(locator.server,
                      "service/rest/v1/search"
                      f"?repository={locator.repository}"
                      f"&name={locator.component_name}"
                      + (f"&group={locator.component_group}" if locator.component_group is not None else ""))

    components = []
    continuation_token = -1  # -1 is a special value hinting the first iteration

    while continuation_token:
        query_url = api_url
        if continuation_token != -1:
            query_url += "&continuationToken=" + continuation_token

        try:
            json_resp = fetch_json(url=query_url, verify=locator.verify_tls_certificate)
        except IOError as e:
            log.error("Cannot fetch components. "
                      "Are you sure that the server URL '%s' is correct "
                      "and the repository '%s' actually exists? Exception is: %s",
                      locator.server, locator.repository, e)
            return []

        continuation_token = json_resp["continuationToken"]
        components += [_parse_component(json_component) for json_component in json_resp["items"]]

    return components


def _parse_component(json_component: dict) -> ComponentPrototype:
    nexus_version = json_component["version"]
    version = nexus_version

    # Special case for Maven repositories:
    # For snapshots, Nexus claims that the version contains a timestamp.
    # This code recovers the true version.
    if json_component["format"] == "maven2" and len(json_component["assets"]) != 0:
        sample_path = json_component["assets"][0]["path"]
        if "snapshot" in sample_path.lower():
            build_number = version[version.rindex("-") + 1:]
            version = sample_path.split("/")[-2] + "." + build_number

    assets = [_parse_asset(json_asset) for json_asset in json_component["assets"]
              if _is_relevant_asset(json_asset)]

    return ComponentPrototype(version=version, assets=assets,
                              extra_data=Nexus3ComponentData(nexus_version=json_component["version"]))


# Filters out irrelevant checksum and PGP signature files.
def _is_relevant_asset(json_asset: dict) -> bool:
    file_ext = os.path.splitext(json_asset["path"])[1]
    return not any(s in file_ext for s in ["md5", "sha", "asc"])


def _parse_asset(json_asset: dict) -> AssetPrototype:
    filename = os.path.basename(json_asset["path"])

    # Special case for NuGet repositories.
    if json_asset["format"] == "nuget":
        filename = json_asset["path"].replace("/", ".") + ".nupkg"

    return AssetPrototype(filename=filename,
                          file_size=None,
                          url=json_asset["downloadUrl"],
                          checksums=json_asset["checksum"])



import logging
from typing import List
from urllib.parse import urljoin

from socorepo.consts import CLF_SOURCES
from socorepo.locators.helpers import fetch_json
from socorepo.locators.pypi import PyPI
from socorepo.structs import ComponentPrototype, AssetPrototype

log = logging.getLogger("socorepo")


def fetch_component_prototypes(locator: PyPI) -> List[ComponentPrototype]:
    if not locator.verify_tls_certificate:
        log.warning("As specified in the repository settings, will now connect to the HTTPS server '%s' without"
                    " verifying its TLS certificate.", locator.server)

    query_url = urljoin(locator.server, f"{locator.project}/json")

    try:
        json_resp = fetch_json(url=query_url, verify=locator.verify_tls_certificate)
    except IOError as e:
        log.error("Cannot fetch components. "
                  "Are you sure that the server URL '%s' is correct "
                  "and the project '%s' actually exists? Exception is: %s",
                  locator.server, locator.project, e)
        return []

    return [ComponentPrototype(version=version,
                               assets=[_parse_asset(json_asset) for json_asset in json_assets])
            for version, json_assets in json_resp["releases"].items()]


def _parse_asset(json_asset: dict) -> AssetPrototype:
    forced_clfs = []
    if json_asset["packagetype"] == "sdist":
        forced_clfs.append(CLF_SOURCES)

    return AssetPrototype(filename=json_asset["filename"],
                          file_size=json_asset["size"],
                          url=json_asset["url"],
                          checksums=json_asset["digests"],
                          forced_clfs=forced_clfs)



from flask_wtf import FlaskForm
from wtforms import TextField, SelectField, SubmitField

from socorepo import config


def create_component_filter_form(available_qualifiers):
    default_qualifier = config.DEFAULT_VERSION_QUALIFIER

    class ComponentFilterForm(FlaskForm):
        version = TextField("Version (starts with):")
        qualifier = SelectField("Qualifier:",
                                choices=[("", "Any")] + [(q.name, q.name) for q in available_qualifiers],
                                default=default_qualifier.name if default_qualifier in available_qualifiers else "")
        filter = SubmitField("Filter")

    return ComponentFilterForm



import os
import shutil


def default_config():
    src_dir = os.path.join(os.path.dirname(__file__), "../socorepo/default_config")
    dst_dir = os.path.abspath("config")

    if os.path.exists(dst_dir):
        print(f"Destination directory '{dst_dir}' already exists. Please delete it and then try again.")
        return

    print(f"Extracting default Socorepo config files to '{dst_dir}'...")
    shutil.copytree(src_dir, "config")
    print("Done!")



from __future__ import annotations

import logging
import os
import re
import sys
from typing import List, Type, Pattern

import toml

log = logging.getLogger("socorepo")


class TomlDict:
    _filename: str
    _dict: dict
    _prefix: List[str]

    @staticmethod
    def load(file: str) -> TomlDict:
        return TomlDict(filename=os.path.basename(file), dict_=toml.load(file), prefix=[])

    def __init__(self, filename: str, dict_: dict, prefix: List[str]):
        self._filename = filename
        self._dict = dict_
        self._prefix = prefix

    def sub(self, key: str, *, split_key: bool = True) -> TomlDict:
        try:
            return self._sub(key.split(".") if split_key else [key])
        except LookupError as e:
            log.error(str(e))
            sys.exit()

    def req(self, key: str, type_: Type, *, split_key: bool = True, choices: list = None):
        try:
            return self._get(key, type_, split_key, choices)
        except LookupError as e:
            log.error(str(e))
            sys.exit()

    def opt(self, key: str, type_: Type, *, split_key: bool = True, choices: list = None, apply: callable = None,
            fallback=None):
        try:
            val = self._get(key, type_, split_key, choices)
            return apply(val) if apply else val
        except LookupError:
            return fallback

    def merge(self, other: TomlDict):
        for k, v2 in other._dict.items():
            if k not in self._dict:
                self._dict[k] = v2
            else:
                v1 = self._dict[k]
                if isinstance(v1, dict) and isinstance(v2, dict):
                    self.sub(k).merge(other.sub(k))
                elif isinstance(v1, list) and isinstance(v2, list):
                    v1 += v2

    def apply_templates(self, available_templates: TomlDict):
        while "__templates__" in self._dict:
            template_id = self._dict["__templates__"].pop(0)
            if template_id not in available_templates._dict:
                log.error("%s: Included template '%s' not found in '%s'.",
                          self.error_prefix, template_id, available_templates._filename)
                sys.exit()

            self.merge(available_templates.sub(template_id))
            if len(self._dict["__templates__"]) == 0:
                del self._dict["__templates__"]

    @property
    def error_prefix(self):
        loc = f"In config file '{self._filename}'"
        if self._prefix:
            loc += f" under key prefix '" + ".".join(self._prefix) + "'"
        return loc

    def __iter__(self):
        return iter(self._dict)

    def _sub(self, key_arr: List[str]):
        # Check that the key exists.
        if key_arr[0] not in self._dict:
            raise LookupError(f"{self.error_prefix}: Expected key '{key_arr[0]}' is missing.")

        val = self._dict[key_arr[0]]

        # Check that the value is a dict.
        if not isinstance(val, dict):
            raise LookupError(f"{self.error_prefix}: Expected key '{key_arr[0]}' to be a table/dict. "
                              f"Found a value of type '{type(val).__name__}' instead.")

        next_dict = TomlDict(filename=self._filename, dict_=val, prefix=self._prefix + [key_arr[0]])
        return next_dict if len(key_arr) == 1 else next_dict._sub(key_arr[1:])

    def _get(self, key: str, type_: Type, split_key: bool, choices: list):
        key_arr = key.split(".") if split_key else [key]

        # This IF determines whether we've arrived at the last segment of the composite key.
        if len(key_arr) == 1:  # Note that in this case: key == key_arr[0]
            # Check that the key exists.
            if key not in self._dict:
                raise LookupError(f"{self.error_prefix}: Expected key '{key}' is missing.")

            val = self._dict[key]
            val = self._parse_and_check_value(key, type_, choices, val)
            return val
        else:
            return self._sub(key_arr[:-1])._get(key_arr[-1], type_, False, choices)

    def _parse_and_check_value(self, key: str, type_: type, choices: list, val):
        if type_ == Pattern:
            # If the programmer requested a regex pattern, we try to parse it for him...
            if not isinstance(val, str):
                raise LookupError(f"{self.error_prefix}: Expected value behind key '{key}' to be a regex string. "
                                  f"Found a value of type '{type(val).__name__}' instead.")
            try:
                # Note that ALL regexes are considered to be case-insensitive!
                val = re.compile(val, re.IGNORECASE)
            except re.error as e:
                raise LookupError(f"{self.error_prefix}: Error while parsing regex behind key '{key}': {e}")
        else:
            # ... and otherwise, we just check whether the value is of the type requested by the programmer.
            # The actual type conversion has already been done internally by TOML.
            if not isinstance(val, type_):
                raise LookupError(f"{self.error_prefix}: Expected value behind key '{key}' to be of type "
                                  f"'{type_.__name__}'. Found a value of type '{type(val).__name__}' instead.")

        # Make sure that the value is not a dict, because in that case, the programmer would need to use sub().
        if isinstance(val, dict):
            raise LookupError(f"{self.error_prefix}: You cannot retrieve table/dict value behind key '{key}' "
                              "with req() or opt(). Use sub() instead.")

        # If the programmer specified a limited list of choices for the value,
        # make sure the value is one of those choices.
        if choices is not None and val not in choices:
            msg_choices = ", ".join(f"'{choice}'" for choice in choices)
            raise LookupError(f"{self.error_prefix}: Expected value behind key '{key}' to be one of {msg_choices}. "
                              f"Found value '{val}' instead.")

        # In case we modified the value, return the new one.
        return val



from typing import List, Dict

from markupsafe import Markup

from socorepo.config.toml_dict import TomlDict
from socorepo.structs import Project, VersionQualifier, AssetClfMatcher

EXTERNAL_CONFIG: bool

LOG_DIR: str
APPLICATION_ROOT: str
FETCH_INTERVAL: int

APPEARANCE_TITLE: str
APPEARANCE_HEADING: str
APPEARANCE_FAVICON_PATH: str
APPEARANCE_HOMEPAGE: Markup
APPEARANCE_FOOTER: Markup

VERSION_QUALIFIERS: List[VersionQualifier]
DEFAULT_VERSION_QUALIFIER: VersionQualifier

ASSET_CLFS: List[str]
ASSET_CLF_MATCHERS: List[AssetClfMatcher]

PROJECTS: Dict[str, Project]



import os
import re

from setuptools import setup, find_packages

DIR = os.path.dirname(__file__)

with open(os.path.join(DIR, "socorepo", "__init__.py"), encoding="utf-8") as f:
    version_file = f.read()
version_match = re.search(r'^__version__ = "([^"]*)"', version_file, re.M)
if not version_match:
    raise RuntimeError("Unable to find version string.")
VERSION = version_match.group(1)

with open(os.path.join(DIR, "README.md"), encoding="utf-8") as f:
    # Skip the first two lines since they are just the heading and look weird on PyPI.
    next(f)
    next(f)
    README = f.read()

setup(
    name="socorepo",
    version=VERSION,
    description="A webapp that fetches software version listings and presents them as download sites.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/LoadingByte/socorepo",
    author="LoadingByte",
    author_email="hello@loadingbyte.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Internet"
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=[
        "toml",
        "requests",
        "Flask",
        "Flask-WTF"
    ],
    entry_points={
        "console_scripts": [
            "socorepo-default-config=socorepo_cli:default_config"
        ]
    }
)



import os
import sys
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler
from zipfile import ZipFile, ZIP_LZMA


def setup_logging(log_dir):
    log_cfg = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "default": {
                "format": "%(asctime)s %(name)-8s %(levelname)-8s %(message)s"
            }
        },
        "handlers": {
            "console": {
                "level": "INFO",
                "formatter": "default",
                "class": "logging.StreamHandler",
                "stream": sys.stderr
            }
        },
        "loggers": {
            "": {  # root
                "handlers": ["console"],
                "level": "NOTSET"
            },
            "socorepo": {
                "level": "INFO"
            }
        }
    }

    if log_dir:
        os.makedirs(log_dir, exist_ok=True)
        log_cfg["handlers"]["file"] = {
            "level": "INFO",
            "formatter": "default",
            "()": _rotating_file_handler,
            "filename": os.path.join(log_dir, "socorepo.log"),
            "maxBytes": 512 * 1024,
            "backupCount": 5,
        }
        log_cfg["loggers"][""]["handlers"].append("file")

    dictConfig(log_cfg)


def _rotating_file_handler(filename, maxBytes, backupCount):
    handler = RotatingFileHandler(filename, maxBytes=maxBytes, backupCount=backupCount)
    handler.namer = _logrot_namer
    handler.rotator = _logrot_rotator
    return handler


def _logrot_namer(name):
    return name + ".zip"


def _logrot_rotator(source, dest):
    with ZipFile(dest, "w", ZIP_LZMA) as df:
        df.write(source, os.path.basename(source))
    os.remove(source)



from __future__ import annotations

from dataclasses import dataclass
from typing import Union, List, Dict
from urllib.parse import urljoin

from markupsafe import Markup

from socorepo.config.toml_dict import TomlDict
from socorepo.structs import Locator, ComponentPrototype, Component


@dataclass(frozen=True)
class GitHubTags(Locator):
    owner: str
    repository: str
    auth: Union[GitHubBasicAuth, GitHubOAuthToken, GitHubOAuthApp]

    def fetch_component_prototypes(self) -> List[ComponentPrototype]:
        from . import fetcher  # avoid cyclic dependencies
        return fetcher.fetch_component_prototypes(self)

    def component_info_table(self, component: Component) -> Dict[str, Markup]:
        owner_url = urljoin("https://github.com", self.owner)
        repository_url = owner_url + "/" + self.repository
        tag_url = repository_url + "/releases/tag/" + component.version
        commit_url = repository_url + "/commit/" + component.extra_data.commit

        return {
            "Component host": Markup("GitHub repository with version tags"),
            "Repository": Markup(
                f'<a href="{owner_url}" target="_blank">{self.owner}</a> / '
                f'<a href="{repository_url}" target="_blank">{self.repository}</a>',
            ),
            "Tag / Commit": Markup(
                f'<a href="{tag_url}" target="_blank">{component.version}</a> / '
                f'<a href="{commit_url}" target="_blank">{component.extra_data.commit:.7}</a>'
            )
        }


@dataclass(frozen=True)
class GitHubBasicAuth:
    username: str
    password: str


@dataclass(frozen=True)
class GitHubOAuthToken:
    oauth_token: str


@dataclass(frozen=True)
class GitHubOAuthApp:
    client_id: str
    client_secret: str


@dataclass(frozen=True)
class GitHubTagsComponentData:
    commit: str


def parse_locator(toml_locator: TomlDict):
    auth = None

    if "login" in toml_locator:
        auth = GitHubBasicAuth(username=toml_locator.req("login.username", str),
                               password=toml_locator.req("login.password", str))
    elif "oauth_token" in toml_locator:
        auth = GitHubOAuthToken(oauth_token=toml_locator.req("oauth_token", str))
    elif "oauth_app" in toml_locator:
        auth = GitHubOAuthApp(client_id=toml_locator.req("oauth_app.client_id", str),
                              client_secret=toml_locator.req("oauth_app.client_secret", str))

    return GitHubTags(owner=toml_locator.req("owner", str),
                      repository=toml_locator.req("repository", str),
                      auth=auth)
